! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                      Block for data arrays                   !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Module for a single block containing several 4D data arrays.
!!
!! A block is a rectangular subset of the spatial grid.
!! It is the smallest unit for load balancing.
!!
!! Blocks have an additional field $weight which respects their computational weight.
!! Load balancing routines (should) use this weight.
!! The default value of the weight is the number of grid cells of the block.
!!
!!
!! Preprocessor options:
!!  - $FD4_COMM_XFER_WGHT: also transfer current block weight to new destination
!!
!!
!! *TODO*:
!!  - do we need diagonal neighbor pointers?
!!  - Test with contigous array for all blocks (equal sized blocks only)
!!
!! @author Matthias Lieber
module fd4_block_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_mpi_mod
  use rbtree_int_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH

  ! public types
  public :: fd4_block
  public :: fd4_data_link
  public :: fd4_block_link
  public :: fd4_cell_link
  public :: fd4_r_k_link
  public :: rbtree_int
  
  ! public methods
  public :: fd4_block_allocate
  public :: fd4_block_deallocate
  public :: fd4_block_empty
  public :: fd4_block_reset
  
  
  !! mask for empty boundary planes flag ($empty)
  !! do not change, fd4_balance relies on this values!
  integer(i4k), parameter, public :: BLOCK_EMPTY(3,2) = reshape( (/2, 4, 8, 16, 32, 64/), (/3,2/) )

  !! A helper to build an array of pointers to data fields. 
  type fd4_data_link
    real(r_k), pointer :: l(:,:,:,:) => null()
  end type fd4_data_link

  !! A helper to build an array of pointers to blocks.
  type fd4_block_link
    type(fd4_block), pointer :: l => null()
  end type fd4_block_link

  !! A helper to build an array of pointers to cells.
  type fd4_cell_link
    real(r_k), pointer :: l(:) => null()
  end type fd4_cell_link
  
  !! A helper used to build an array of pointers
  !! (not used in $fd4_block, but elsewhere)
  type fd4_r_k_link
    real(r_k), pointer :: l => NULL()
  end type fd4_r_k_link


  !! Derived Type for a block.
  !!
  !! The actual data fields are allocated in $sdata:
  !!  - $sdata(0)%l  ... all cell-centered variables
  !!  - $sdata(1)%l  ... all x-face variables
  !!  - $sdata(2)%l  ... all y-face variables
  !!  - $sdata(3)%l  ... all z-face variables
  !! The field bounds are: ( vars * timesteps_per_var * bins_per_var , x, y, z)
  !!
  !! *CAUTION*: $fields contains pointers to non-continuous array sections of $sdata(0:3)%l.
  !! Be careful to avoid creation of temporary arrays!
  !!
  !! *TODO*:
  !!  - fd4_block_empty: diagonal and trigonal neighbors
  !!  - added ext to $fd4_block -> *utilize this in all routines which need the extents*
  !!  - do we need dynamic face vars?
  !!
  !! @see fd4_vartab_mod
  type fd4_block
    integer(i_k) :: pos(3)                                !! position of the block in the block decomposition
    integer(i_k) :: bnd(3,2)                              !! bounds of the block in global grid coordinates
    integer(i_k) :: ext(3)                                !! extents of the block
    type(fd4_block_link) :: neigh(3,2)                    !! array of pointers to spatial neighbors of this block

    !! array of pointers to array sections of $sdata, representing the data fields
    type(fd4_data_link), pointer :: fields(:,:) => null()

    type(fd4_data_link) :: sdata(0:4)                     !! storage arrays for static data fields (allocated at block creation)
    
    integer(i4k), pointer :: empty(:,:) => null()         !! tells which fields and boundaries are empty (< $vthres of vartab)
    
    integer(i4k) :: sempty                                !! summarized empty info (all vars and time steps) of me

    real(r4k) :: weight                                   !! computational weight of the block

    integer :: bltype = MPI_DATATYPE_NULL                 !! MPI datatype for this block
    integer(MPI_ADDRESS_KIND) :: bladdr = 0               !! address of block%empty(1,1)

  end type fd4_block

contains  
  

  !! Allocates all of the static arrays of the block.
  !!
  !! *The block itself must be allocated*! For internal use only.
  !! Error checks require level 3.
  subroutine fd4_block_allocate(b, bext, vartab, varinfo, init, build_type, alloc_face, err)
    implicit none
    type(fd4_block),   pointer      :: b           !! the block to allocate
    integer(i_k),      intent(in)   :: bext(3)     !! spatial extent of the block                         
    type(fd4_vartab),  intent(in)   :: vartab(:)   !! variable table containing the definitions of fields to allocate             
    type(fd4_varinfo), intent(in)   :: varinfo     !! variable table info
    logical,           intent(in)   :: init        !! init data fields with vnull?
    logical,           intent(in)   :: build_type  !! built an MPI type for the block?
    logical,           intent(in)   :: alloc_face  !! allocate face variables?
    integer(i_k),      intent(out)  :: err         !! error status, 0 on sucess                                                   

    integer(i_k) :: i, st, len(0:3)
    
    err = -1
    
    FD4_L3_ASSERT(size(vartab)<1,'size(vartab)<1')
    
    b%ext(:) = bext(:)
    len(:) = varinfo%field_len(:)
    
    ! allocate "empty"-flags
    allocate( b%empty(size(vartab), varinfo%maxnsteps) , stat=err )
    FD4_L3_ASSERT(err/=0,'allocate b%empty')

    ! allocate data field pointers
    allocate( b%fields(size(vartab), varinfo%maxnsteps) , stat=err )
    FD4_L3_ASSERT(err/=0,'allocate b%fields')
    
    ! allocate static data memory for cell-centered vars
    if( len(0) > 0 ) then
      allocate( b%sdata(0)%l(len(0), bext(1), bext(2), bext(3)) , stat=err )
      FD4_L3_ASSERT(err/=0,'allocate b%sdata(0)%l')
    else
      b%sdata(0)%l => null()
    end if

    ! allocate static data memory for x face vars
    if( alloc_face .and. len(1) > 0 ) then
      allocate( b%sdata(1)%l(len(1), bext(1)+1, bext(2), bext(3)) , stat=err )
      FD4_L3_ASSERT(err/=0,'allocate b%sdata(1)%l')
    else
      b%sdata(1)%l => null()
    end if

    ! allocate static data memory for y face vars
    if( alloc_face .and. len(2) > 0 ) then
      allocate( b%sdata(2)%l(len(2), bext(1), bext(2)+1, bext(3)) , stat=err )
      FD4_L3_ASSERT(err/=0,'allocate b%sdata(2)%l')
    else
      b%sdata(2)%l => null()
    end if

    ! allocate static data memory for z face vars
    if( alloc_face .and. len(3) > 0 ) then
      allocate( b%sdata(3)%l(len(3), bext(1), bext(2), bext(3)+1) , stat=err )
      FD4_L3_ASSERT(err/=0,'allocate b%sdata(3)%l')
    else
      b%sdata(3)%l => null()
    end if

    ! set field pointers
    ! change order of the loops in fdv_vartab_create_varinfo to change the
    ! order of (variable, step) in the array
    
    ! loop over maximum number of time steps
    do st=1,varinfo%maxnsteps
      ! loop over variables
      do i=1,size(vartab)
      
        if( varinfo%fields(i,st,1) > 0 .and. associated(b%sdata(vartab(i)%facevar)%l) ) then
          ! vartab(i) is static data field with >= st time steps
          b%fields(i,st)%l => b%sdata( vartab(i)%facevar )%l( varinfo%fields(i,st,1):varinfo%fields(i,st,2) ,:,:,:)
        else
          ! vartab(i) is dynamic data field or has < st time steps
          b%fields(i,st)%l => null()
        end if
        
      end do
    end do
    
    ! initialize data and not constant metadata 
    call fd4_block_reset(b, vartab, varinfo, init)

    ! build a type for this block
    if(build_type) then
      call fd4_block_build_type(b, err)
      FD4_L3_ASSERT(err/=0,'fd4_block_build_type failed')
    end if

    err = 0

  end subroutine fd4_block_allocate
  


  !! Reset data fields of the block with initial values.
  !! Block must exist.
  !!
  subroutine fd4_block_reset(b, vartab, varinfo, init)
    type(fd4_block),   pointer      :: b           !! the block to allocate
    type(fd4_vartab),  intent(in)   :: vartab(:)   !! variable table containing the definitions of fields to allocate             
    type(fd4_varinfo), intent(in)   :: varinfo     !! variable table info
    logical,           intent(in)   :: init        !! if false: do not init data fields (faster)

    integer(i_k) :: i, st

    ! initialize the weight with some meaningful value:
    !  for now we just take the number of grid cells
    b%weight = REAL( PRODUCT(b%ext) , r4k )

    ! set "empty"-flags
    b%empty(:,:) = 0_i4k
    
    ! shall we set all fields to vnull?
    if(init) then
      if(varinfo%all_vars_same_vnull) then
        ! wow! all vars have the same vnull value, that makes resetting all fields quiet easy & fast:
        !TODO: we could even use this, if only vars of same face type have same vnull
        if(associated(b%sdata(0)%l)) b%sdata(0)%l = vartab(1)%vnull
        if(associated(b%sdata(1)%l)) b%sdata(1)%l = vartab(1)%vnull
        if(associated(b%sdata(3)%l)) b%sdata(2)%l = vartab(1)%vnull
        if(associated(b%sdata(4)%l)) b%sdata(3)%l = vartab(1)%vnull
      else
        ! ok, we have to take the clumsy way
        ! loop over maximum number of time steps
        do st=1,varinfo%maxnsteps
          ! loop over variables
          do i=1,size(vartab)
            if( associated(b%fields(i,st)%l) ) then
              b%fields(i,st)%l = vartab(i)%vnull
            end if
          end do
        end do
      end if
    end if
    
  end subroutine fd4_block_reset
  

  
  !! Deallocate all arrays of a block *and also the block itself*.
  !!
  !! Nullifies the block pointer. For internal use only.
  subroutine fd4_block_deallocate(b)
    implicit none
    type(fd4_block), pointer :: b          !! the block to deallocate
    integer :: err
    
    if(associated(b%empty))  deallocate(b%empty)
    if(associated(b%fields)) deallocate(b%fields)
    if(associated(b%sdata(0)%l))  deallocate(b%sdata(0)%l)
    if(associated(b%sdata(1)%l))  deallocate(b%sdata(1)%l)
    if(associated(b%sdata(2)%l))  deallocate(b%sdata(2)%l)
    if(associated(b%sdata(3)%l))  deallocate(b%sdata(3)%l)
    
    if(b%bltype/=MPI_DATATYPE_NULL) then
      call MPI_Type_free(b%bltype, err)
    end if

    deallocate(b)
    b => null()
      
  end subroutine fd4_block_deallocate
 

  !! Check if a field is empty.
  !!
  !! Fills block component $empty of variable $idx at time step $st.
  !! For internal use only.
  !!
  !! - b%empty(idx,st) == 0  ->  block is completely empty
  !! - b%empty(idx,st) == 1  ->  block is empty at all boundary planes (i.e. requires no neighbors)
  !! - b%empty(idx,st) == 3  ->  block is empty at all boundary planes except for x-1
  !! - ...
  !! see constant module variable $BLOCK_EMPTY.
  !!
  !! *TODO*:
  !!  - optimize for overlapping boundary planes (think of blocksize 2x2x2 and emptyplanes=2!)
  !!  - face vars??
  subroutine fd4_block_empty(b, vartab, varinfo, idx, st, emptyplanes)
    implicit none
    type(fd4_block),  pointer       :: b              !! the block
    type(fd4_vartab), intent(in)    :: vartab(:)      !! variable table
    type(fd4_varinfo),intent(in)    :: varinfo        !! variable info
    integer(i_k),     intent(in)    :: idx            !! field index in vartab
    integer(i_k),     intent(in)    :: st             !! time step index
    !! number of "null" cell planes required to the phase boundary, i.e. to *not* create new block
    integer(i_k),     intent(in)    :: emptyplanes(3)
    
    integer(i_k) :: dir, dim, bin, x, y, z, bext(3), lbnd(0:3,2)

    
    b%empty(idx,st) = 0_i4k
    
    bext = b%ext

    lbnd(1:3,1) = 1
    lbnd(1:3,2) = bext(:)

    lbnd(0,1) = varinfo%fields(idx,st,1)  ! bin start index (fixed)
    lbnd(0,2) = varinfo%fields(idx,st,2)  ! bin end index (fixed)
    
#ifdef FD4_BLOCK_EMPTY_OPT
    if(ALL(bext==emptyplanes)) then
    
      ! block has same size as emptyplanes in all dimensions
      call fd4_block_empty_dataloop(lbnd,127)
      
      
    else
#endif
    
      ! loop over boundary planes
      do dim=1,3
        if(bext(dim)==emptyplanes(dim)) then
          ! block has same size as emptyplanes
          lbnd(dim,1) = 1
          lbnd(dim,2) = bext(dim)
          call fd4_block_empty_dataloop(lbnd,BLOCK_EMPTY(dim,1)+BLOCK_EMPTY(dim,2))
        else
          do dir=1,2
            if(dir==1) then
              lbnd(dim,1) = 1
              lbnd(dim,2) = emptyplanes(dim)
            else
              lbnd(dim,1) = bext(dim) - emptyplanes(dim) + 1
              lbnd(dim,2) = bext(dim)
            end if
            ! check if this boundary plane is empty
            call fd4_block_empty_dataloop(lbnd,BLOCK_EMPTY(dim,dir))        
          end do
        end if
        lbnd(dim,1) = 1
        lbnd(dim,2) = bext(dim)
      end do


      if(b%empty(idx,st) == 0_i4k) then
        ! empty at the boundary planes, check inner block cells
        lbnd(1:3,1) = emptyplanes(:) + 1
        lbnd(1:3,2) = bext(:) - emptyplanes(:)
        ! check if lbnd is not null (does not improve performance)
        if(lbnd(1,1)<=lbnd(1,2) .and. &
           lbnd(2,1)<=lbnd(2,2) .and. &
           lbnd(3,1)<=lbnd(3,2) )     &
           call fd4_block_empty_dataloop(lbnd, 1)      
      else
        b%empty(idx,st) = IOR(b%empty(idx,st), 1)
      end if

#ifdef FD4_BLOCK_EMPTY_OPT    
    end if
#endif
    
    ! if (b%empty(idx,st) == 0_i4k)   -> block is completely empty
    ! if (b%empty(idx,st) == 1_i4k)   -> block is empty at all boundary planes (i.e. requires no neighbors)
    
    ! print *,b%pos,b%empty(idx,st)

  contains
  
    subroutine fd4_block_empty_dataloop(lbnd, empty)
      integer(i_k), intent(in) :: lbnd(0:3,2)
      integer(i4k), intent(in) :: empty
      dataloop: &
      do z=lbnd(3,1),lbnd(3,2)
        do y=lbnd(2,1),lbnd(2,2)
          do x=lbnd(1,1),lbnd(1,2)
            do bin=lbnd(0,1),lbnd(0,2)
              if( abs(b%sdata(0)%l(bin,x,y,z)) > vartab(idx)%vthres ) then
                b%empty(idx,st) = IOR(b%empty(idx,st), empty)
                exit dataloop
              end if
            end do
          end do
        end do
      end do dataloop
    end subroutine fd4_block_empty_dataloop
    
  end subroutine fd4_block_empty  


  
  !! Routine to build an MPI datatype to migrate the block.
  !!
  !! Build an MPI datatype for the $fd4_block components:
  !! - $sempty
  !! - $empty
  !! - $weight (only if compiled with $FD4_COMM_XFER_WGHT)
  !! - $sdata(0:3)%l
  !!
  !! The resulting data type is not commited!
  !! Don't forget to MPI_Type_free the type after use.
  !!
  !! Internal routine.
  subroutine fd4_block_build_type(b, err)
    type(fd4_block),  pointer                        :: b           !! the block
    integer(i_k), intent(out)                        :: err         !! error status: 0...ok

    integer(i_k) :: i
    integer :: bl_count
    integer :: bl_types(7)
#ifdef FD4_COMM_XFER_WGHT
    parameter ( bl_types = (/ MPI_I4K, MPI_I4K, MPI_R4K, MPI_R_K, MPI_R_K, MPI_R_K, MPI_R_K /) )
#else
    parameter ( bl_types = (/ MPI_I4K, MPI_I4K, MPI_R_K, MPI_R_K, MPI_R_K, MPI_R_K, 0 /) )
#endif
    integer :: bl_blens(size(bl_types))
    integer(MPI_ADDRESS_KIND) :: bl_displ(size(bl_types))
    err = -1
    bl_blens = 0
    bl_count = 0

    ! displacements and counts of sempty and empty
    bl_count = bl_count + 1
    call MPI_Get_address( b%sempty, bl_displ(bl_count), err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Get_address failed')
    bl_blens(bl_count) = 1

    bl_count = bl_count + 1
    call MPI_Get_address( b%empty(1,1), bl_displ(bl_count), err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Get_address failed')
    bl_blens(bl_count) = size( b%empty )


#ifdef FD4_COMM_XFER_WGHT
    ! also transfer current block weight to new destination
    bl_count = bl_count + 1
    call MPI_Get_address( b%weight, bl_displ(bl_count), err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Get_address failed')
    bl_blens(bl_count) = 1
#endif

    ! displacements and counts of data arrays
    do i=0,3
      ! check if the data array is allocated
      if(associated(b%sdata(i)%l)) then
        bl_count = bl_count + 1
        call MPI_Get_address( b%sdata(i)%l(1,1,1,1), bl_displ(bl_count), err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Get_address failed')
        bl_blens(bl_count) = size( b%sdata(i)%l )
      end if
    end do
    
    b%bladdr = bl_displ(1)

    do i=2,bl_count
      bl_displ(i) = bl_displ(i) - bl_displ(1)
    end do
    bl_displ(1) = 0

    call MPI_Type_create_struct(bl_count, bl_blens, bl_displ, bl_types, b%bltype, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Type_create_struct failed')
    
    err = 0
  
  end subroutine fd4_block_build_type


end module fd4_block_mod
