! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                  Boundary condition handling                 !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Apply boundary conditions on a domain.
!!
!! Currently we have zero gradient and a simple method to set specific values.
!!
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_boundary_mod

  use fd4_domain_mod
  use fd4_block_mod
  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_ghostlist_mod
  
  implicit none
  private
  
  public :: fd4_boundary_zerograd
  public :: fd4_boundary_zerograd_block
  public :: fd4_boundary_spec
  public :: fd4_boundary_spec_block

contains

  !! Apply zero gradient boundary conditions on all blocks of a domain.
  !!
  !! Sets the given cell-centered vars in the boundary ghost block to the same value as
  !! in the cells at domain boundary.
  !! Sets the value of the given face variables at domain boundary (within domain) to vnull.
  !!
  !! Valid values for $opt_dim are $FD4_X, $FD4_Y, $FD4_Z, $FD4_XY, $FD4_XZ, $FD4_YZ, and $FD4_XYZ.
  !! If no dimensions are specified, than boundary conditions for all 3 spatial dimensions are set
  !! ($FD4_XYZ).
  !!
  !! If $opt_dir is not given or is not 1 or 2, both lower and upper boundaries are set.
  !!
  subroutine fd4_boundary_zerograd(domain, idx, st, opt_dim, opt_dir)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    integer(i_k)    , intent(in)    :: idx(:)        !! array of field indices
    integer(i_k)    , intent(in)    :: st(:)         !! array of time step indices
    integer(i_k)    , intent(in)    :: opt_dim       !! optional specification of spatial dimensions
    integer(i_k)    , intent(in)    :: opt_dir       !! optional specification of boundary direction (1...lower, 2...upper)
    optional :: opt_dim, opt_dir
    
    type(fd4_block) , pointer       :: bgb    ! the boundary ghost block
    type(fd4_block) , pointer       :: b      ! the (ordinary) block at domain boundary
    integer(i_k) :: dim, dir

#if ( FD4_VERBOSE_LEVEL >= 3 )
      integer(i_k) :: f
      ! error checks
      do f=1,MIN(size(idx),size(st))
        if(idx(f)<0 .or. idx(f)>ubound(domain%vartab,1)) then
          FD4_L1_ERROR('field idx not valid f='//i2s(f))
        end if
        if(st(f)<0 .or. st(f)>domain%vartab(idx(f))%nsteps) then
          FD4_L1_ERROR('time step not valid f='//i2s(f))
        end if
      end do
#endif

    dim = FD4_XYZ
    if(present(opt_dim)) dim = opt_dim
    
    dir=3
    if(present(opt_dir)) then
      dir = opt_dir
      if(dir<1 .or. dir>3) dir = 3
    end if

    ! will be null if we have periodic bounds or no blocks at boundary
    bgb => domain%glist(FD4_GHOST_BND)%start
    
    ! loop over all boundary ghost blocks
    do while(associated(bgb))
      
      b => bgb%neigh(2,1)%l  ! the boundary block within the domain
      
      ! check if dim contains the spatial dimensions of this block
      if( IAND( FD4_DIM(bgb%pos(3)), dim ) /= 0 .and. IAND( bgb%pos(2), dir ) /= 0) then
        call fd4_boundary_zerograd_set(domain, b, bgb, idx, st)
      end if
    
      bgb => bgb%neigh(1,2)%l  ! iterate to next
    end do
  
  end subroutine fd4_boundary_zerograd




  !! Apply zero gradient boundary conditions on specified block
  !!
  !! Sets the given cell-centered vars in the boundary ghost block to the same value as
  !! in the cells at domain boundary.
  !! Sets the value of the given face variables at domain boundary (within domain) to vnull.
  !!
  !! Valid values for $opt_dim are $FD4_X, $FD4_Y, $FD4_Z, $FD4_XY, $FD4_XZ, $FD4_YZ, and $FD4_XYZ.
  !! If no dimensions are specified, than boundary conditions for all 3 spatial dimensions are set
  !! ($FD4_XYZ).
  !!
  !! If $opt_dir is not given or is not 1 or 2, both lower and upper boundaries are set.
  !!
  subroutine fd4_boundary_zerograd_block(domain, b, idx, st, opt_dim, opt_dir)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(fd4_block) , pointer       :: b             !! the block
    integer(i_k)    , intent(in)    :: idx(:)        !! array of field indices
    integer(i_k)    , intent(in)    :: st(:)         !! array of time step indices
    integer(i_k)    , intent(in)    :: opt_dim       !! optional specification of spatial dimensions
    integer(i_k)    , intent(in)    :: opt_dir       !! optional specification of boundary direction (1...lower, 2...upper)
    optional :: opt_dim, opt_dir
    
    integer(i_k) :: dim, dir

#if ( FD4_VERBOSE_LEVEL >= 3 )
      integer(i_k) :: f
      ! error checks
      do f=1,MIN(size(idx),size(st))
        if(idx(f)<0 .or. idx(f)>ubound(domain%vartab,1)) then
          FD4_L1_ERROR('field idx not valid f='//i2s(f))
        end if
        if(st(f)<0 .or. st(f)>domain%vartab(idx(f))%nsteps) then
          FD4_L1_ERROR('time step not valid f='//i2s(f))
        end if
      end do
#endif

    ! will be null if we have periodic bounds or no blocks at boundary
    if(.not. associated(domain%glist(FD4_GHOST_BND)%start)) return
    
    dir=3
    if(present(opt_dir)) then
      dir = opt_dir
      if(dir<1 .or. dir>3) dir = 3
    end if

    ! identify boundary ghost blocks
    do dim=1,3
      if(domain%periodic(dim)) cycle
      
      ! check if opt_dim contains the spatial direction of this block
      if(present(opt_dim)) then
        if( IAND( FD4_DIM(dim), opt_dim ) == 0 ) cycle
      end if
      
      ! lower boundary
      if(b%pos(dim)==1 .and. IAND(dir,1)/=0) then
        call fd4_boundary_zerograd_set(domain, b, b%neigh(dim,1)%l, idx, st)
      end if
      
      ! upper boundary
      if(b%pos(dim)==domain%blnum(dim) .and. IAND(dir,2)/=0) then
        call fd4_boundary_zerograd_set(domain, b, b%neigh(dim,2)%l, idx, st)
      end if

    end do
    
  end subroutine fd4_boundary_zerograd_block


  
  !! Apply zero gradient boundary conditions on specified block/boundary ghost block pair.
  !!
  !! Internal subroutine. Used by $fd4_boundary_zerograd and $fd4_boundary_zerograd_block.
  subroutine fd4_boundary_zerograd_set(domain, b, bgb, idx, st)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(fd4_block) , pointer       :: b             !! the block
    type(fd4_block) , pointer       :: bgb           !! the boundary ghost block
    integer(i_k)    , intent(in)    :: idx(:)        !! array of field indices
    integer(i_k)    , intent(in)    :: st(:)         !! array of time step indices
    
    integer(i_k) :: sc, sf, g, f, nfields
    
    nfields = MIN(size(idx),size(st))

    ! bgb%pos(1) = boundary ghost block indicator (must be 1)
    ! bgb%pos(3) = boundary ghost block dimension relative to normal block (x,y,z resp. 1,2,3)
    ! bgb%pos(2) = boundary ghost block direction relative to normal block ( -1,1 resp. 1,2)
    
    if(bgb%pos(2)==1) then
      sc = 1                 ! bgb is before b, lower slice is copied   (cell centered)
      sf = 1                 ! bgb is before b, lower slice is set to 0 (facevar)
    else
      sc = b%ext(bgb%pos(3)) ! bgb is behind b, upper slice is copied   (cell centered)
      sf = sc + 1            ! bgb is behind b, upper slice is set to 0 (facevar)
    end if

    ! x
    if(bgb%pos(3)==1) then      
      do f=1,nfields
        if( domain%vartab(idx(f))%facevar == 1 ) then
          b%fields(idx(f),st(f))%l(:,sf,:,:) = domain%vartab(idx(f))%vnull
        else if( domain%vartab(idx(f))%facevar == 0 ) then        
          do g=1,domain%nghosts(1)
            bgb%fields(idx(f),st(f))%l(:,g,:,:) = b%fields(idx(f),st(f))%l(:,sc,:,:)
          end do
        end if
      end do
    end if

    ! y
    if(bgb%pos(3)==2) then      
      do f=1,nfields
        if( domain%vartab(idx(f))%facevar == 2 ) then
          b%fields(idx(f),st(f))%l(:,:,sf,:) = domain%vartab(idx(f))%vnull
        else if( domain%vartab(idx(f))%facevar == 0 ) then
          do g=1,domain%nghosts(2)
            bgb%fields(idx(f),st(f))%l(:,:,g,:) = b%fields(idx(f),st(f))%l(:,:,sc,:)
          end do
        end if
      end do
    end if

    ! z
    if(bgb%pos(3)==3) then      
      do f=1,nfields
        if( domain%vartab(idx(f))%facevar == 3 ) then
          b%fields(idx(f),st(f))%l(:,:,:,sf) = domain%vartab(idx(f))%vnull
        else if( domain%vartab(idx(f))%facevar == 0 ) then
          do g=1,domain%nghosts(3)
            bgb%fields(idx(f),st(f))%l(:,:,:,g) = b%fields(idx(f),st(f))%l(:,:,:,sc)
          end do
        end if
      end do
    end if
  
  end subroutine fd4_boundary_zerograd_set






  !! Apply specific boundary conditions on all blocks of a domain.
  !!
  !! $values(1) is used for the first boundary layer, $values(2) for the second etc.
  !! Does only process cell centered variables. Only one spatial dimension and direction
  !! (upper, lower) is processed. Sets all variables to the same value.
  !!
  subroutine fd4_boundary_spec(domain, idx, st, face, dir, values)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    integer(i_k)    , intent(in)    :: idx(:)        !! array of field indices
    integer(i_k)    , intent(in)    :: st(:)         !! array of time step indices
    integer(i_k)    , intent(in)    :: face          !! dimension to process (1,2,3)
    integer(i_k)    , intent(in)    :: dir           !! direction to process (1...lower face, 2... upper face)
    real(r_k)       , intent(in)    :: values(:)     !! values for each ghost plane
    
    type(fd4_block) , pointer       :: bgb    ! the boundary ghost block
#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: f
#endif

    FD4_L1_ASSERT(dir<1.or.dir>2,'dir must be 1 or 2')
    FD4_L1_ASSERT(face<1.or.face>3,'face must be 1, 2, or 3')    
    FD4_L1_ASSERT(size(values)<domain%nghosts(face),'too few values')

#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      do f=1,MIN(size(idx),size(st))
        if(idx(f)<0 .or. idx(f)>ubound(domain%vartab,1)) then
          FD4_L1_ERROR('field idx not valid f='//i2s(f))
        end if
        if(st(f)<0 .or. st(f)>domain%vartab(idx(f))%nsteps) then
          FD4_L1_ERROR('time step not valid f='//i2s(f))
        end if
        if(domain%vartab(idx(f))%facevar /= 0) then
          FD4_L1_ERROR('not a cell centered variable f='//i2s(f))
        end if
      end do
#endif

    ! will be null if we have periodic bounds or no blocks at boundary
    bgb => domain%glist(FD4_GHOST_BND)%start
    
    ! loop over all boundary ghost blocks
    do while(associated(bgb))
      
      if(bgb%pos(3)==face .and. bgb%pos(2)==dir) then
        call fd4_boundary_spec_set(domain, bgb, idx, st, values)
      end if
    
      bgb => bgb%neigh(1,2)%l  ! iterate to next
    end do
  
  end subroutine fd4_boundary_spec



  !! Apply specific boundary conditions to a single block $b of a domain.
  !!
  !! $values(1) is used for the first boundary layer, $values(2) for the second etc.
  !! Does only process cell centered variables. Only one spatial dimension and direction
  !! (upper, lower) is processed. Sets all variables to the same value.
  !!
  subroutine fd4_boundary_spec_block(domain, b, idx, st, face, dir, values)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(fd4_block) , pointer       :: b             !! the block
    integer(i_k)    , intent(in)    :: idx(:)        !! array of field indices
    integer(i_k)    , intent(in)    :: st(:)         !! array of time step indices
    integer(i_k)    , intent(in)    :: face          !! dimension to process (1,2,3)
    integer(i_k)    , intent(in)    :: dir           !! direction to process (1...lower face, 2... upper face)
    real(r_k)       , intent(in)    :: values(:)     !! values for each ghost plane

#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: f
#endif

    FD4_L1_ASSERT(dir<1.or.dir>2,'dir must be 1 or 2')
    FD4_L1_ASSERT(face<1.or.face>3,'face must be 1, 2, or 3')    
    FD4_L1_ASSERT(size(values)<domain%nghosts(face),'too few values')

#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! error checks
      do f=1,MIN(size(idx),size(st))
        if(idx(f)<0 .or. idx(f)>ubound(domain%vartab,1)) then
          FD4_L1_ERROR('field idx not valid f='//i2s(f))
        end if
        if(st(f)<0 .or. st(f)>domain%vartab(idx(f))%nsteps) then
          FD4_L1_ERROR('time step not valid f='//i2s(f))
        end if
        if(domain%vartab(idx(f))%facevar /= 0) then
          FD4_L1_ERROR('not a cell centered variable f='//i2s(f))
        end if
      end do
#endif

    ! will be null if we have periodic bounds or no blocks at boundary
    if(.not. associated(domain%glist(FD4_GHOST_BND)%start)) return

    if(domain%periodic(face)) return

    ! lower boundary
    if(b%pos(face)==1 .and. dir==1) then
      FD4_L3_ASSERT(.not.associated(b%neigh(face,1)%l),'bgb not present')
      call fd4_boundary_spec_set(domain, b%neigh(face,1)%l, idx, st, values)
    end if

    ! upper boundary
    if(b%pos(face)==domain%blnum(face) .and. dir==2) then
      FD4_L3_ASSERT(.not.associated(b%neigh(face,2)%l),'bgb not present')
      call fd4_boundary_spec_set(domain, b%neigh(face,2)%l, idx, st, values)
    end if

  end subroutine fd4_boundary_spec_block



  !! Apply specific boundary conditions on specified block/boundary ghost block pair.
  !!
  !! Internal subroutine. Used by $fd4_boundary_spec.
  subroutine fd4_boundary_spec_set(domain, bgb, idx, st, values)
    implicit none
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(fd4_block) , pointer       :: bgb           !! the boundary ghost block
    integer(i_k)    , intent(in)    :: idx(:)        !! array of field indices
    integer(i_k)    , intent(in)    :: st(:)         !! array of time step indices
    real(r_k),target, intent(in)    :: values(:)     !! values for each ghost plane
    
    integer(i_k) :: g, f, nfields
    
    nfields = MIN(size(idx),size(st))

    ! bgb%pos(1) = boundary ghost block indicator (must be 1)
    ! bgb%pos(3) = boundary ghost block dimension relative to normal block (x,y,z resp. 1,2,3)
    ! bgb%pos(2) = boundary ghost block direction relative to normal block ( -1,1 resp. 1,2)
    
    ! x
    if(bgb%pos(3)==1) then      
      do f=1,nfields
        do g=1,domain%nghosts(1)
          bgb%fields(idx(f),st(f))%l(:,g,:,:) = values(g)
        end do
      end do
    end if

    ! y
    if(bgb%pos(3)==2) then      
      do f=1,nfields
        do g=1,domain%nghosts(2)
          bgb%fields(idx(f),st(f))%l(:,:,g,:) = values(g)
        end do
      end do
    end if

    ! z
    if(bgb%pos(3)==3) then      
      do f=1,nfields
        do g=1,domain%nghosts(3)
          bgb%fields(idx(f),st(f))%l(:,:,:,g) = values(g)
        end do
      end do
    end if
  
  end subroutine fd4_boundary_spec_set




end module fd4_boundary_mod
